/*
 * Copyright (C) 2011 Giorgio Vazzana
 *
 * This file is part of hwebcam.
 *
 * hwebcam is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <signal.h>
#include <poll.h>
#include <time.h>
#include <sys/time.h>
#include <theora/theoraenc.h>
#include <vorbis/codec.h>
#include <vorbis/vorbisenc.h>
#include "video.h"
#include "audio.h"

#ifdef USE_SDL
#include "SDL.h"
#endif

#define VERSION "0.3"
#define DEBUG   0

struct ppm_st {
	unsigned int   width;
	unsigned int   height;
	unsigned int   max;
	unsigned char *data;
};

enum muxer_page_type {
	AUDIO,
	VIDEO
};

struct muxer_page_st {
	ogg_page              opage;
	struct muxer_page_st *next;
};

struct muxer_st {
	FILE                 *fpvideo;

	th_enc_ctx           *tctx;
	struct muxer_page_st *vq_head;
	struct muxer_page_st *vq_tail;

	int                   haveaudio;
	vorbis_dsp_state     *vds;
	struct muxer_page_st *aq_head;
	struct muxer_page_st *aq_tail;
};

int keep_going = 1;

void signal_handler(int sig)
{
	if (sig == SIGINT)
		keep_going = 0;
}

#define CLIPRGB(x) if(x < 0) x = 0; else if(x > 255) x = 255;
#define COMPUTE_RGB(y, cb, cr) r = (256 * y +          + 359 * cr + 128) >> 8;\
                               g = (256 * y -  88 * cb - 183 * cr + 128) >> 8;\
                               b = (256 * y + 454 * cb +          + 128) >> 8;
void yuyv422_to_rgb24(const unsigned char *src, struct ppm_st *frame)
{
	unsigned int i, j;
	int y1, cb, y2, cr;
	int r, g, b;

	for (i = 0; i < frame->height; i++) {
		for (j = 0; j < frame->width; j += 2) {
			y1 = *(src++);
			cb = *(src++) - 128;
			y2 = *(src++);
			cr = *(src++) - 128;

			COMPUTE_RGB(y1, cb, cr)
			CLIPRGB(r)
			CLIPRGB(g)
			CLIPRGB(b)

			*(frame->data + 3*j + 3*i*frame->width + 0) = r;
			*(frame->data + 3*j + 3*i*frame->width + 1) = g;
			*(frame->data + 3*j + 3*i*frame->width + 2) = b;

			COMPUTE_RGB(y2, cb, cr)
			CLIPRGB(r)
			CLIPRGB(g)
			CLIPRGB(b)

			*(frame->data + 3*(j+1) + 3*i*frame->width + 0) = r;
			*(frame->data + 3*(j+1) + 3*i*frame->width + 1) = g;
			*(frame->data + 3*(j+1) + 3*i*frame->width + 2) = b;
		}
	}
}

void yuyv422_to_yuv420p(const unsigned char *src, th_ycbcr_buffer ycbcrbuf)
{
	int i, j, y1, cb, y2, cr;
	int two_rows_y, i_rows_y, i_rows_cb, i_rows_cr;

	two_rows_y = 2 * ycbcrbuf[0].width;
	for (i = 0; i < ycbcrbuf[0].height; i++) {
		i_rows_y  =      i*ycbcrbuf[0].width;
		i_rows_cb = (i>>1)*ycbcrbuf[1].width;
		i_rows_cr = (i>>1)*ycbcrbuf[2].width;
		for (j = 0; j < ycbcrbuf[0].width; j += 2) {
			y1 = src[0];
			cb = src[1];
			y2 = src[2];
			cr = src[3];

			*(ycbcrbuf[0].data + i_rows_y + j  ) = y1;
			*(ycbcrbuf[0].data + i_rows_y + j+1) = y2;

			if (!(i&1)) {
				cb += src[two_rows_y + 1];
				*(ycbcrbuf[1].data + i_rows_cb + (j>>1)) = cb >> 1;

				cr += src[two_rows_y + 3];
				*(ycbcrbuf[2].data + i_rows_cr + (j>>1)) = cr >> 1;
			}
			src += 4;
		}
	}
}

void muxer_add_page(struct muxer_st *muxer, ogg_page *opage, enum muxer_page_type type)
{
	struct muxer_page_st **tail, *oldtail;

	if (type == AUDIO)
		tail = &muxer->aq_tail;
	else
		tail = &muxer->vq_tail;

	oldtail = *tail;

	if (*tail == NULL) {
		*tail = malloc(sizeof(**tail));
	} else {
		(*tail)->next = malloc(sizeof(**tail));
		*tail = (*tail)->next;
	}

	if (*tail) {
		(*tail)->opage = *opage;
		(*tail)->next  = NULL;
		(*tail)->opage.header = malloc(opage->header_len);
		(*tail)->opage.body   = malloc(opage->body_len);

		if ((*tail)->opage.header && (*tail)->opage.body) {
			memcpy((*tail)->opage.header, opage->header, opage->header_len);
			memcpy((*tail)->opage.body,   opage->body,   opage->body_len);
		} else {
			free((*tail)->opage.header);
			free((*tail)->opage.body);
			free(*tail);
			*tail = oldtail;
		}
	} else
		*tail = oldtail;

	if (oldtail == NULL) {
		if (type == AUDIO)
			muxer->aq_head = *tail;
		else
			muxer->vq_head = *tail;
	}
}

void muxer_write(struct muxer_st *muxer, int flush)
{
	ogg_int64_t granulepos_theora;
	ogg_int64_t granulepos_vorbis;
	double      time_theora;
	double      time_vorbis;
	struct muxer_page_st *p;

	/* if we have only a video stream we can just flush pages */
	if (muxer->haveaudio == 0)
		flush = 1;

	while (1) {
		/* if both the queues are empty, there's nothing to do */
		if (muxer->aq_head == NULL && muxer->vq_head == NULL)
			break;

		/* look for pages with granulepos != -1 in audio queue */
		granulepos_vorbis = -1;
		p = muxer->aq_head;
		while (p) {
			granulepos_vorbis = ogg_page_granulepos(&p->opage);
			if (granulepos_vorbis != -1)
				break;
			p = p->next;
		}

		/* look for pages with granulepos != -1 in video queue */
		granulepos_theora = -1;
		p = muxer->vq_head;
		while (p) {
			granulepos_theora = ogg_page_granulepos(&p->opage);
			if (granulepos_theora != -1)
				break;
			p = p->next;
		}

		/* if one of the queue is empty or doesn't have pages with
		   granulepos != -1, then we're done */
		if (!flush && (granulepos_vorbis == -1 || granulepos_theora == -1))
			break;

		/* ok, at this point both granulepos are != -1, we can compute times */
		time_vorbis = vorbis_granule_time(muxer->vds, granulepos_vorbis);
		time_theora = th_granule_time(muxer->tctx, granulepos_theora);

		/* write audio pages */
		if (time_vorbis <= time_theora || flush) {
			while (1) {
				if (muxer->aq_head == NULL)
					break;

				p = muxer->aq_head;
				muxer->aq_head = muxer->aq_head->next;
				if (muxer->aq_head == NULL)
					muxer->aq_tail = NULL;

				granulepos_vorbis = ogg_page_granulepos(&p->opage);
				fwrite(p->opage.header, 1, p->opage.header_len, muxer->fpvideo);
				fwrite(p->opage.body  , 1, p->opage.body_len,   muxer->fpvideo);
				free(p->opage.header);
				free(p->opage.body);
				free(p);

				/* stop when we find a page with granulepos != -1 */
				if (!flush && granulepos_vorbis != -1)
					break;
			}
		}

		/* write video pages */
		if (time_vorbis > time_theora || flush) {
			while (1) {
				if (muxer->vq_head == NULL)
					break;

				p = muxer->vq_head;
				muxer->vq_head = muxer->vq_head->next;
				if (muxer->vq_head == NULL)
					muxer->vq_tail = NULL;

				granulepos_theora = ogg_page_granulepos(&p->opage);
				fwrite(p->opage.header, 1, p->opage.header_len, muxer->fpvideo);
				fwrite(p->opage.body  , 1, p->opage.body_len,   muxer->fpvideo);
				free(p->opage.header);
				free(p->opage.body);
				free(p);

				if (!flush && granulepos_theora != -1)
					break;
			}
		}
	}
}

void usage(void)
{
	fprintf(stderr,
"usage: hwebcam [options]\n"
"       -h          print help\n"
"       -v          increment verbosity level\n"
"Video options:\n"
"       -d videodev use video device videodev      (/dev/video0)\n"
"       -s WxH      choose frame size              (640x400)\n"
"       -i fps      choose frame rate              (10)\n"
"       -t val      save a frame every val seconds (disabled, 0=every frame)\n"
"       -e          encode captured stream in ogg/theora/vorbis\n"
"       -o ofile    save captured stream in ofile  (capture.ogg, -=stdout)\n"
"       -b bitrate  choose theora bitrate in bit/s (200000)\n"
"       -q quality  choose theora quality          (0, range=[0-63])\n"
#ifdef USE_SDL
"       -w          display stream in a sdl window\n"
#endif
"Audio options:\n"
"       -a          include audio in captured stream\n"
"       -D audiodev use audio device audiodev      (hw:0,0)\n"
"       -c channels choose number of channels      (1)\n"
"       -r rate     choose sampling rate           (32000)\n"
"       -B bitrate  choose vorbis bitrate in bit/s (48000)\n"
"       -Q quality  choose vorbis quality          (0.1, range=[-0.1 - 1.0])\n"
	);

	exit(1);
}

int main(int argc, char *argv[])
{
	int ret, opt, i, j, verbose = 0;
	struct sigaction   act;
	struct pollfd      pfds[4];
	int                nfds, pollret;
	struct timeval     tp0, tp1;   /* poll */
	struct timeval     tea0, tea1; /* encode audio */
	struct timeval     tev0, tev1; /* encode video */
	long int           t_poll, t_enca, t_encv;

	/* video data */
	struct video_data  vd;
	char              *videodevice = "/dev/video0";
	int                width = 640, height = 400, fps = 10;
	struct timeval     frame_start, frame_stop;
	time_t             timeout_start, timeout_stop;
	long               frame_interval;
	double             actual_fps;

	/* audio data */
	struct audio_data  ad;
	char              *audiodevice = "hw:0,0";
	unsigned int       rate = 32000, channels = 1, peak_count = 0;
	float              peak_percent = 0.0f;

	/* image data */
	struct ppm_st      ppmimage;
	int                timeout = -1;
	FILE              *fpimage = NULL;
	char               imagename[32] = "image.ppm";

	/* ogg data */
	ogg_stream_state   ostream_theora;
	ogg_stream_state   ostream_vorbis;
	ogg_page           opage;
	FILE              *fpvideo = NULL;
	char              *videoname = "capture.ogg";
	struct muxer_st    muxer;

	/* theora data */
	int                encode_theora = 0;
	th_info            ti;
	th_comment         tc;
	th_enc_ctx        *tctx = NULL;
	th_ycbcr_buffer    ycbcrbuf;
	ogg_packet         opacket_theora;
	int                bitrate_theora = 200000, quality_theora = 0;
	int                duplicate = 0, keep_going_theora = 0;

	/* vorbis data */
	int                encode_vorbis = 0;
	vorbis_info        vi;
	vorbis_comment     vc;
	vorbis_dsp_state   vds;
	vorbis_block       vb;
	ogg_packet         opacket_vorbis[3];
	float            **vorbisbuffer;
	unsigned int       c, s;
	int                bitrate_vorbis = 48000, keep_going_vorbis = 0;
	float              quality_vorbis = 0.1f;

#ifdef USE_SDL
	int                sdl_window = 0;
	char               caption_string[32];
	SDL_Surface       *surface = NULL;
	SDL_Overlay       *overlay = NULL;
	SDL_Rect           video_rect;
#endif

	fprintf(stderr, "hwebcam, ver. %s\n\n", VERSION);

#ifdef USE_SDL
	while ((opt = getopt(argc, argv, "hvd:s:i:t:eo:b:q:waD:c:r:B:Q:")) != -1) {
#else
	while ((opt = getopt(argc, argv, "hvd:s:i:t:eo:b:q:aD:c:r:B:Q:")) != -1) {
#endif
		switch (opt) {
		case 'h':
			usage();
			break;
		case 'v':
			verbose++;
			break;
		case 'd':
			videodevice = optarg;
			break;
		case 's':
			ret = sscanf(optarg, "%dx%d", &width, &height);
			if (ret != 2) {
				fprintf(stderr, "Error: cannot read frame size\n\n");
				usage();
			}
			break;
		case 'i':
			fps = atoi(optarg);
			break;
		case 't':
			timeout = atoi(optarg);
			break;
		case 'e':
			encode_theora = 1;
			break;
		case 'o':
			videoname = optarg;
			break;
		case 'b':
			bitrate_theora = atoi(optarg);
			break;
		case 'q':
			quality_theora = atoi(optarg);
			bitrate_theora = 0;
			break;
#ifdef USE_SDL
		case 'w':
			sdl_window = 1;
			break;
#endif
		case 'a':
			encode_vorbis = 1;
			break;
		case 'D':
			audiodevice = optarg;
			break;
		case 'c':
			channels = atoi(optarg);
			break;
		case 'r':
			rate = atoi(optarg);
			break;
		case 'B':
			bitrate_vorbis = atoi(optarg);
			break;
		case 'Q':
			quality_vorbis = atof(optarg);
			bitrate_vorbis = 0;
			break;
		default:
			fprintf(stderr, "\n");
			usage();
		}
	}

#if DEBUG
	fprintf(stderr,
	        "v=%d.  Video: d=%s, s=%dx%d, i=%d, t=%d, e=%d, o=%s, b=%d, q=%d, w=%d."
	        "  Audio: a=%d, D=%s, c=%u, r=%u, B=%d, Q=%.2f\n",
	        verbose, videodevice, width, height, fps, timeout, encode_theora,
	        videoname, bitrate_theora, quality_theora, sdl_window, encode_vorbis,
	        audiodevice, channels, rate, bitrate_vorbis, quality_vorbis);
#endif

	if (width<0 || width>1280 || height<0 || height>1024 || fps<0 || fps>30 ||
	    bitrate_theora<0 || quality_theora<0 || quality_theora>63 ||
	    channels>6 || rate>96000 || bitrate_vorbis<0 || quality_vorbis<-0.1f ||
	    quality_vorbis>1.0f) {
		fprintf(stderr, "Error: parameters out of bounds\n\n");
		usage();
	}
	/* we can only do: theora, theora+vorbis. we don't do vorbis alone */
	if (encode_vorbis == 1 && encode_theora == 0)
		encode_vorbis = 0;

	/* Install signal handler */
	act.sa_handler = signal_handler;
	sigemptyset(&act.sa_mask);
	act.sa_flags = SA_RESTART;
	ret = sigaction(SIGINT, &act, NULL);
	if (ret == -1)  {
		fprintf(stderr, "Error: sigaction() failed\n");
		exit(1);
	}

	/* Init video */
	ret = video_init(&vd, videodevice, width, height, fps, verbose >= 2 ? 1 : 0);
	if (ret == -1)
		exit(1);
	if ((unsigned int)width != vd.width || (unsigned int)height != vd.height ||
	    (unsigned int)fps != vd.tpfden) {
		fprintf(stderr, "Warning: video device do not support %dx%d@%d, %dx%d@%d chosen instead\n",
				width, height, fps, vd.width, vd.height, vd.tpfden);
		width  = vd.width;
		height = vd.height;
		fps    = vd.tpfden;
	}
	/* file descriptor for video is always at pfds[0] */
	pfds[0].fd     = vd.fd;
	pfds[0].events = POLLIN;
	nfds = 1;
	/* if we are encoding in theora, width and height must be multiple of 16 */
	if (encode_theora && (width % 16 + height % 16)) {
		fprintf(stderr, "Error: image dimensions are not multiple of 16\n");
		exit(1);
	}

	/* Init alsa audio */
	if (encode_vorbis) {
		ret = audio_init(&ad, audiodevice, rate, channels, pfds+1, verbose >= 2 ? 1 : 0);
		if (ret == -1)
			exit(1);
		nfds += ad.nfds;
	}

	/* Prepare ppm image */
	if (timeout >= 0) {
		ppmimage.width  = width;
		ppmimage.height = height;
		ppmimage.max    = 255;
		ppmimage.data   = malloc(3 * ppmimage.width * ppmimage.height * sizeof(unsigned char));
		if (ppmimage.data == NULL)  {
			fprintf(stderr, "Error: malloc() failed\n");
			exit(1);
		}
	}

	/* Prepare theora and, if necessary, vorbis encoders */
	if (encode_theora) {
		/* open output file */
		if (!strcmp(videoname, "-"))
			fpvideo = stdout;
		else {
			fpvideo = fopen(videoname, "w");
			if (fpvideo == NULL) {
				fprintf(stderr, "Error: cannot open %s\n", videoname);
				exit(1);
			}
		}

		/* prepare ogg theora stream with random serialno */
		srand(time(NULL));
		ret = ogg_stream_init(&ostream_theora, rand());
		if (ret != 0) {
			fprintf(stderr, "Error: ogg_stream_init() failed\n");
			exit(1);
		}

		/* configure theora encoder */
		th_info_init(&ti);
		th_comment_init(&tc);
		th_comment_add(&tc, "ENCODER=hwebcam");

		ti.frame_width            = width;
		ti.frame_height           = height;
		ti.pic_width              = width;
		ti.pic_height             = height;
		ti.pic_x                  = 0;
		ti.pic_y                  = 0;
		ti.colorspace             = TH_CS_UNSPECIFIED;
		ti.pixel_fmt              = TH_PF_420;
		ti.target_bitrate         = bitrate_theora;
		ti.quality                = quality_theora;
		ti.keyframe_granule_shift = 6;
		ti.version_major          = 0;
		ti.version_minor          = 0;
		ti.version_subminor       = 0;
		ti.fps_numerator          = fps;
		ti.fps_denominator        = 1;
		ti.aspect_numerator       = 1;
		ti.aspect_denominator     = 1;

		/* allocate theora encoder handle */
		tctx = th_encode_alloc(&ti);
		if (tctx == NULL) {
			fprintf(stderr, "Error: th_encode_alloc() failed\n");
			exit(1);
		}

		/* We need to push the packets/pages in this order:
		   1st       theora header packet, in his own page
		   1st       vorbis header packet, in his own page    (if present)
		   remaining vorbis headers, flush the page when done (if present)
		   remaining theora headers, flush the page when done  */

		/* 1st theora header packet */
		if ((ret = th_encode_flushheader(tctx, &tc, &opacket_theora)) > 0) {
			if (ogg_stream_packetin(&ostream_theora, &opacket_theora) == 0) {
				while (ogg_stream_pageout(&ostream_theora, &opage)) {
					fwrite(opage.header, 1,  opage.header_len, fpvideo);
					fwrite(opage.body,   1,  opage.body_len,   fpvideo);
				}
			} else {
				fprintf(stderr, "Error: ogg_stream_packein() failed\n");
				exit(1);
			}
		}
		if (ret == TH_EFAULT) {
			fprintf(stderr, "Error: th_encode_flushheader() failed\n");
			exit(1);
		}


		if (encode_vorbis) {
			/* prepare ogg vorbis stream with random serialno */
			ret = ogg_stream_init(&ostream_vorbis, rand());
			if (ret != 0) {
				fprintf(stderr, "Error: ogg_stream_init() failed\n");
				exit(1);
			}

			/* configure vorbis and initialize encoder */
			vorbis_info_init(&vi);
			if (bitrate_vorbis == 0)
				ret = vorbis_encode_init_vbr(&vi, ad.channels, ad.rate, quality_vorbis);
			else
				ret = vorbis_encode_init(&vi, ad.channels, ad.rate, -1, bitrate_vorbis, -1);
			if (ret != 0) {
				fprintf(stderr, "Error: vorbis_encode_init*() failed\n");
				exit(1);
			}
			vorbis_analysis_init(&vds, &vi);
			vorbis_comment_init(&vc);
			vorbis_comment_add(&vc, "ENCODER=hwebcam");

			/* vorbis header packets */
			ret = vorbis_analysis_headerout(&vds, &vc,
			      &opacket_vorbis[0], &opacket_vorbis[1], &opacket_vorbis[2]);
			if (ret != 0) {
				fprintf(stderr, "Error: vorbis_analysis_headerout() failed\n");
				exit(1);
			}
			for (i = 0; i < 3; i++)  {
				if (ogg_stream_packetin(&ostream_vorbis, &opacket_vorbis[i]) == 0) {
					while (ogg_stream_pageout(&ostream_vorbis, &opage)) {
						fwrite(opage.header, 1, opage.header_len, fpvideo);
						fwrite(opage.body  , 1, opage.body_len,   fpvideo);
					}
				} else {
					fprintf(stderr, "Error: ogg_stream_packein() failed\n");
					exit(1);
				}
			}

			/* force remaining vorbis header packets into a page */
			while (ogg_stream_flush(&ostream_vorbis, &opage)) {
				fwrite(opage.header, 1, opage.header_len, fpvideo);
				fwrite(opage.body,   1, opage.body_len,   fpvideo);
			}

			vorbis_block_init(&vds, &vb);

			/* ok, vorbis encoder is ready */
			keep_going_vorbis = 1;
		}

		/* remaining theora header packets */
		while ((ret = th_encode_flushheader(tctx, &tc, &opacket_theora)) > 0) {
			if (ogg_stream_packetin(&ostream_theora, &opacket_theora) == 0) {
				while (ogg_stream_pageout(&ostream_theora, &opage)) {
					fwrite(opage.header, 1,  opage.header_len, fpvideo);
					fwrite(opage.body,   1,  opage.body_len,   fpvideo);
				}
			} else {
				fprintf(stderr, "Error: ogg_stream_packein() failed\n");
				exit(1);
			}
		}
		if (ret == TH_EFAULT) {
			fprintf(stderr, "Error: th_encode_flushheader() failed\n");
			exit(1);
		}
		/* force remaining theora header packets into a page */
		while (ogg_stream_flush(&ostream_theora, &opage)) {
			fwrite(opage.header,  1, opage.header_len, fpvideo);
			fwrite(opage.body,    1, opage.body_len,   fpvideo);
		}

		/* ok, theora encoder is ready */
		keep_going_theora = 1;

		/* now, set up theora's ycbcr buffer */
		ycbcrbuf[0].width  = width;
		ycbcrbuf[0].height = height;
		ycbcrbuf[0].stride = ycbcrbuf[0].width;
		ycbcrbuf[0].data   = malloc(ycbcrbuf[0].width * ycbcrbuf[0].stride * sizeof(unsigned char));

		ycbcrbuf[1].width  = width/2;
		ycbcrbuf[1].height = height/2;
		ycbcrbuf[1].stride = ycbcrbuf[1].width;
		ycbcrbuf[1].data   = malloc(ycbcrbuf[1].width * ycbcrbuf[1].stride * sizeof(unsigned char));

		ycbcrbuf[2].width  = width/2;
		ycbcrbuf[2].height = height/2;
		ycbcrbuf[2].stride = ycbcrbuf[2].width;
		ycbcrbuf[2].data   = malloc(ycbcrbuf[2].width * ycbcrbuf[2].stride * sizeof(unsigned char));

		if (!ycbcrbuf[0].data || !ycbcrbuf[1].data || !ycbcrbuf[2].data) {
			fprintf(stderr, "Error: malloc() failed\n");
			exit(1);
		}

		/* prepare muxer */
		muxer.fpvideo   = fpvideo;
		muxer.tctx      = tctx;
		muxer.vq_head   = NULL;
		muxer.vq_tail   = NULL;
		if (encode_vorbis) {
			muxer.haveaudio = 1;
			muxer.vds       = &vds;
		} else {
			muxer.haveaudio = 0;
			muxer.vds       = NULL;
		}
		muxer.aq_head   = NULL;
		muxer.aq_tail   = NULL;
	}

#ifdef USE_SDL
	if (sdl_window) {
		/* Initialize SDL Video subsystem */
		if (verbose)
			fprintf(stderr, "Initializing SDL...\n");
		if (SDL_Init(SDL_INIT_VIDEO) == -1) {
			fprintf(stderr, "Error: could not initialize SDL: %s\n", SDL_GetError());
			exit(1);
		}

		/* Clean up on exit */
		atexit(SDL_Quit);

		/* Initialize the display requesting a software surface */
		surface = SDL_SetVideoMode(width, height, 24, SDL_SWSURFACE|SDL_ANYFORMAT);
		if (surface == NULL) {
			fprintf(stderr, "Error: couldn't set requested video mode: %s\n", SDL_GetError());
			exit(1);
		}
		if (verbose)
			fprintf(stderr, "  Surface: %dx%d at %d bits-per-pixel mode\n",
			        surface->w, surface->h, surface->format->BitsPerPixel);

		/* Create a YUV video overlay  */
		overlay = SDL_CreateYUVOverlay(width, height, SDL_YUY2_OVERLAY, surface);
		if (verbose) {
			fprintf(stderr, "  Overlay:\n");
			fprintf(stderr, "    Dimensions:     %dx%d\n", overlay->w, overlay->h);
			fprintf(stderr, "    Format:         0x%08x\n", overlay->format);
			fprintf(stderr, "    Planes:         %d\n", overlay->planes);
			for (i = 0; i < overlay->planes; i++)
				fprintf(stderr, "    Pitches[%d]:     %d\n", i, overlay->pitches[i]);
			fprintf(stderr, "    Hw accelerated: %d\n", overlay->hw_overlay);
		}

		/* Set the window title and icon name */
		SDL_WM_SetCaption("hwebcam", "hwebcam");
	}
#endif

	/* Print streaming info */
	fprintf(stderr, "Info:\n");
	fprintf(stderr, "  Video device: %s\n", videodevice);
	fprintf(stderr, "  Frame size:   %dx%d\n", width, height);
	fprintf(stderr, "  Frame rate:   %d fps\n", fps);
	if (timeout >= 0)
	fprintf(stderr, "  Save frame:   every %d seconds\n", timeout);
#ifdef USE_SDL
	fprintf(stderr, "  Window:       %s\n", sdl_window ? "SDL" : "none");
#endif
	if (encode_theora) {
	fprintf(stderr, "  Theora stream:\n");
	fprintf(stderr, "    libtheora:  %s\n", th_version_string());
	fprintf(stderr, "    File name:  %s\n", strcmp(videoname, "-") == 0 ? "stdout" : videoname);
	fprintf(stderr, "    Bitrate:    %d bit/s\n", ti.target_bitrate);
	fprintf(stderr, "    Quality:    %d\n", ti.quality);
	}
	if (encode_vorbis) {
	fprintf(stderr, "  Audio device: %s\n", audiodevice);
	fprintf(stderr, "  Vorbis stream:\n");
	fprintf(stderr, "    libvorbis:  %s\n", vorbis_version_string());
	fprintf(stderr, "    Channels:   %d\n", vi.channels);
	fprintf(stderr, "    Rate:       %ld\n", vi.rate);
	if (bitrate_vorbis > 0)
	fprintf(stderr, "    Bitrate:    %d bit/s\n", bitrate_vorbis);
	else
	fprintf(stderr, "    Quality:    %.2f\n", quality_vorbis);
	}

	/* Main loop */
	/* start video capture */
	ret = video_start_streaming(&vd);
	if (ret == -1)
		exit(1);
	/* wait until the first video frame is ready */
	pollret = poll(pfds, 1, -1);

	if (encode_vorbis) {
		/* start audio capture */
		ret = audio_start(&ad);
		if (ret == -1)
			exit(1);
	}

	if (verbose)
		fprintf(stderr, "\n");
	frame_start.tv_sec  = 0;	//silence warning
	frame_start.tv_usec = 0;
	timeout_start = time(NULL);
	t_enca = 0;
	t_encv = 0;
	i = 0;
	while (keep_going || keep_going_theora || keep_going_vorbis) {
		/* poll: is audio or video ready? */
		gettimeofday(&tp0, NULL);
		pollret = poll(pfds, nfds, -1);
		gettimeofday(&tp1, NULL);
		t_poll = (tp1.tv_sec - tp0.tv_sec) * 1000000L + tp1.tv_usec - tp0.tv_usec;

		/* poll can fail and return -1 when a signal arrives. errno will be EINTR */
		if (pollret == -1) {
			//perror("poll");
			continue;
		}

		if (pollret == 0) {
			fprintf(stderr, "Poll timeout\n");
			continue;
		}

		/** AUDIO **/
		if (encode_vorbis) {
			audio_poll_descriptors_revents(&ad);

			if (ad.revents & POLLERR) {
				fprintf(stderr, "Audio POLLERR\n");
				exit(1);
			}

			/* process audio if ready */
			if (ad.revents & POLLIN && keep_going_vorbis) {
				/* read ad.period_size frames of audio */
				ret = audio_update_avail_and_read(&ad);
				if (ret == -1)
					exit(1);

				/* encoding */
				gettimeofday(&tea0, NULL);

				if (peak_count >= 100) {
					peak_count   = 0;
					peak_percent = 0.0f;
				}
				audio_find_peak(&ad);
				if (peak_percent < ad.peak_percent)
					peak_percent = ad.peak_percent;
				peak_count++;

				vorbisbuffer = vorbis_analysis_buffer(&vds, ad.frames);
				for(c = 0; c < ad.channels; c++) {
					for (s = 0; s < ad.frames; s++) {
						vorbisbuffer[c][s] = (float)ad.alsabuffer[s * ad.channels + c] / 32767.0f;
					}
				}
				if (keep_going == 0)
					keep_going_vorbis = 0;

				ret = vorbis_analysis_wrote(&vds, keep_going_vorbis ? ad.frames : 0);
				if (ret != 0) {
					fprintf(stderr, "Error: vorbis_analysis_wrote() failed\n");
					exit(1);
				}

				while ((ret = vorbis_analysis_blockout(&vds, &vb)) > 0) {
					ret = vorbis_analysis(&vb, NULL);
					if (ret != 0) {
						fprintf(stderr, "Error: vorbis_analysis() failed\n");
						exit(1);
					}
					ret = vorbis_bitrate_addblock(&vb);
					if (ret != 0) {
						fprintf(stderr, "Error: vorbis_bitrate_addblock() failed\n");
						exit(1);
					}

					while ((ret = vorbis_bitrate_flushpacket(&vds, &opacket_vorbis[0])) > 0) {
						if (ogg_stream_packetin(&ostream_vorbis, &opacket_vorbis[0]) == 0) {
							while (ogg_stream_pageout(&ostream_vorbis, &opage))
								muxer_add_page(&muxer, &opage, AUDIO);
						} else {
							fprintf(stderr, "Error: ogg_stream_packein() failed\n");
							exit(1);
						}
					}
					if (ret < 0) {
						fprintf(stderr, "Error: vorbis_bitrate_flushpacket() failed\n");
						exit(1);
					}
				}
				if (ret < 0) {
					fprintf(stderr, "Error: vorbis_analysis_blockout() failed\n");
					exit(1);
				}
				muxer_write(&muxer, 0);
				gettimeofday(&tea1, NULL);
				t_enca = (tea1.tv_sec - tea0.tv_sec) * 1000000L + tea1.tv_usec - tea0.tv_usec;
#if DEBUG
				fprintf(stderr, "pr=%d  pt=%7ldus  avail=%5ld  frames=%5ld  tea=%5ldus\n",
				        pollret, t_poll, ad.avail, ad.frames, t_enca);
#endif
			}
		}

		/** VIDEO **/
		if (pfds[0].revents & POLLERR) {
			fprintf(stderr, "Video POLLERR\n");
			exit(1);
		}

		/* process video if ready */
		if (pfds[0].revents & POLLIN) {
			/* get a video frame */
			ret = video_frame_get(&vd);
			if (ret == -1)
				exit(1);

			/* compute frame time and actual fps */
			frame_stop     = vd.buffer.timestamp;
			if (i == 0)
				frame_start= frame_stop;
			frame_interval = (frame_stop.tv_sec - frame_start.tv_sec) * 1000000L +
			                  frame_stop.tv_usec - frame_start.tv_usec;
			actual_fps     = (frame_interval == 0) ? fps : (1000000.0 / (double)frame_interval);
			frame_start    = vd.buffer.timestamp;

#ifdef USE_SDL
			if (sdl_window) {
				/* Fill in pixel data */
				SDL_LockYUVOverlay(overlay);
				for (j = 0; j < height; j++)
					memcpy(overlay->pixels[0]+j*overlay->pitches[0],
					       vd.buffers[vd.buffer.index].start+j*overlay->pitches[0],
					       (size_t) overlay->pitches[0]);
				SDL_UnlockYUVOverlay(overlay);
				/* Display frame */
				video_rect.x = 0;
				video_rect.y = 0;
				video_rect.w = width;
				video_rect.h = height;
				SDL_DisplayYUVOverlay(overlay, &video_rect);
				/* Print fps in window's title, every 1 second */
				if (actual_fps > 0.5 && (i % (int)(actual_fps+0.5)) == 0) {
					sprintf(caption_string, "hwebcam - %5.2f fps", actual_fps);
					SDL_WM_SetCaption(caption_string, "hwebcam");
				}
			}
#endif

			if (encode_theora && keep_going_theora) {
				gettimeofday(&tev0, NULL);
				/* prepare the frame for the encoder */
				yuyv422_to_yuv420p(vd.buffers[vd.buffer.index].start, ycbcrbuf);

				/* if actual_fps==fps/2, push 2 frames into the encoder. if fps/4, push 4 frames */
				duplicate = 0;
				if (0.45 * (double)fps < actual_fps && actual_fps < 0.55 * (double)fps)
					duplicate = 1;
				else if (0.20 * (double)fps < actual_fps && actual_fps < 0.30 * (double)fps)
					duplicate = 3;

				/* set the number of duplicates of the next frame to produce */
				if (duplicate)
					ret = th_encode_ctl(tctx, TH_ENCCTL_SET_DUP_COUNT, &duplicate, sizeof(duplicate));
					if (ret != 0) {
						fprintf(stderr, "Error: th_encode_ctl() failed\n");
						exit(1);
					}

				/* pass the frame to the encoder */
				if (th_encode_ycbcr_in(tctx, ycbcrbuf) != 0) {
					fprintf(stderr, "Error: th_encode_ycbcr_in() failed\n");
					exit(1);
				}

				if (keep_going == 0)
					keep_going_theora = 0;

				/* get the packet out and pass it to the ogg stream */
				while ((ret = th_encode_packetout(tctx, !keep_going_theora, &opacket_theora)) > 0) {
					if (ogg_stream_packetin(&ostream_theora, &opacket_theora) == 0) {
						while (ogg_stream_pageout(&ostream_theora, &opage))
							muxer_add_page(&muxer, &opage, VIDEO);
					} else {
						fprintf(stderr, "Error: ogg_stream_packein() failed\n");
						exit(1);
					}
				}
				if (ret == TH_EFAULT) {
					fprintf(stderr, "Error: th_encode_packetout() failed\n");
					exit(1);
				}
				muxer_write(&muxer, 0);
				gettimeofday(&tev1, NULL);
				t_encv = (tev1.tv_sec - tev0.tv_sec) * 1000000L + tev1.tv_usec - tev0.tv_usec;
#if DEBUG
				fprintf(stderr, "pr=%d  pt=%7ldus  frame: %5d  buffer: %2d  tf: %7ld us  fps: %5.2f  tev=%7ldus\n",
				        pollret, t_poll, i, vd.buffer.index, frame_interval, actual_fps, t_encv);
#endif
			}

			if (timeout >= 0) {
				timeout_stop = time(NULL);
				if (difftime(timeout_stop, timeout_start) >= timeout) {
					/* Dump data to ppm image */
					yuyv422_to_rgb24(vd.buffers[vd.buffer.index].start, &ppmimage);
					//sprintf(imagename, "%08d.ppm", i);
					fpimage = fopen(imagename, "w");
					if (fpimage == NULL)  {
						fprintf(stderr, "Error: cannot open %s\n", imagename);
						exit(1);
					}
					fprintf(fpimage, "P6\n%u %u\n%u\n", ppmimage.width, ppmimage.height, ppmimage.max);
					fwrite(ppmimage.data, 3, ppmimage.width * ppmimage.height, fpimage);
					fclose(fpimage);

					if (verbose)
						fprintf(stderr, "  Image saved\n");
					timeout_start = time(NULL);
				}
			}

#if !DEBUG
			/* print stats in verbose mode every frame if actual_fps<=10, or every 2 frames */
			if (verbose && ((int)(actual_fps+0.5)<=10 || !(i%2))) {
				fprintf(stderr, "\rf: %5d  tf: %6ldus  fps: %5.2f  tev: %6ldus  tea: %5ldus  ap: %5.1f%%",
				        i, frame_interval, actual_fps, t_encv, t_enca, peak_percent);
				fflush(stderr);
			}
#endif

			/* ok, we're done with this frame */
			ret = video_frame_done(&vd);
			if (ret == -1)
				exit(1);

			i++;
		}
	}
	fprintf(stderr, "\n");

	ret = video_close(&vd);

	/* clean up */
	if (encode_theora) {

		muxer_write(&muxer, 1);

		th_encode_free(tctx);
		th_info_clear(&ti);
		th_comment_clear(&tc);
		ogg_stream_clear(&ostream_theora);

		if (encode_vorbis) {
			ret = audio_stop(&ad);
			ret = audio_close(&ad);

			vorbis_info_clear(&vi);
			vorbis_comment_clear(&vc);
			vorbis_block_clear(&vb);
			vorbis_dsp_clear(&vds);

			ogg_stream_clear(&ostream_vorbis);
		}

		if (strcmp(videoname, "-"))
			fclose(fpvideo);
	}

	return 0;
}
